<?php


class DPD_Calculator extends WPRuby_Abstract_Calculator
{

	public function __construct($settings)
	{
		$this->settings = $settings;
	}


	/**
	 * @return array
	 */
	public function calculate($cart)
	{

		$packages = $this->get_package_details_by_boxpacker($cart);

		if (!is_array($packages)) {
			return [];
		}

		return $this->get_price_for_packages($packages, $cart);
	}


	private function get_price_for_packages($packages, $cart)
	{

		$parcels = [];
		foreach ($packages as $package) {
			$parcels[] = [
				"Weight" => $package['weight'],
				"Length" => $package['length'],
				"Width" =>  $package['width'],
				"Height" => $package['height'],
			];
		}

		$countries = include ROYALMAILPRO_DIR . 'shared/data/countries.php';
		$countryName = WC()->countries->get_countries()[$cart['destination']['country']];
		$normalizedCountryName = $this->normalizeCountryName($countryName);

		if (!isset($countries[$normalizedCountryName])) {
			return [];
		}

		$countryRates = require ROYALMAILPRO_DIR . 'dpd/data/rates/'. intval($countries[$normalizedCountryName]) .'.php';

		if (!$countryRates) {
			return [];
		}

		$quotes = [];
		foreach ($parcels as $parcel) {
			$weight = ceil($parcel['Weight'] * 2) / 2;
			foreach ($countryRates as $countryRate) {
				if ($countryRate['weight'] == $weight) {
					$quotes[] = $countryRate;
				}
			}
		}
		if (!$quotes) {
			return [];
		}

		$domestic_options = (!is_array($this->settings['domestic_options'])) ? [$this->settings['domestic_options'] ] : $this->settings['domestic_options'];
		$intl_options = (!is_array($this->settings['intl_options'])) ? [$this->settings['intl_options'] ] : $this->settings['intl_options'];

		$enabledServices = array_merge($domestic_options, $intl_options);

		$totalCover = 0;
		if ($this->settings['add_cover'] === 'yes') {
			$totalCover = $this->getTotalCoverFee($packages);
		}

		$rates = [];
		foreach ($quotes as $quote) {
			$serviceSlug = $quote['serviceSlug'];
			if (in_array($serviceSlug, $enabledServices)) {
				$rates[] = [
					'id'    => $serviceSlug,
					'label' => sprintf('%s: %s', $this->settings['title'], $quote['label']),
					'cost'  => $quote['cost'] + $totalCover,
				];
			}

		}

		return $rates;

	}

	private function getTotalCoverFee($packages)
	{
		$total = 0;

		if (!is_array($packages)) {
			return $total;
		}

		foreach ($packages as $package) {
			$total += $this->calculateCoverFees($package['value']);
		}

		return $total;
	}

	/**
	 * @param float $value
	 * @return float
	 */
	private function calculateCoverFees($value)
	{
		if ($value <= 50) {
			return 0;
		}

		if ($value >= 5000 ) {
			return 247.50;
		}

		$valueToCover = $value - 50;

		return round($valueToCover * 0.05, 2);
	}

}
