<?php

/**
 * Class Calculate_Shipping_Price
 */
class WPRuby_Evri_Domestic_Calculator
{

    /**
     * @var array
     */
    private $categories = [];

    /**
     * @var array
     */
    private $cover_rates = [];

    /**
     * @var int
     */
    private $shippingAmount = [];

    /**
     * @var int
     */
    private $orderAmount = 0;


    /**
     * @var int
     */
    private $weight = 0;

    /**
     * @var int
     */
    const SIGNATURE_FEE = 0.90;

	/**
	 * Calculate_Shipping_Price constructor.
	 *
	 * @param $categories
	 * @param $cover_rates
	 */
    public function __construct($categories, $cover_rates)
    {
        $this->cover_rates = $cover_rates;
        $this->categories = $categories;
    }

    /**
     * @param $products
     * @param array $defaults
     * @return $this
     */

    public function add_items($products, $defaults = [])
    {

        $default_weight = (isset($defaults['height'])) ? $defaults['height'] : 1;

        foreach ($products as $product) {
            $_product = $product['data'];
            $this->weight += wc_get_weight(($_product->get_weight() <= 0) ? $default_weight : $_product->get_weight(), 'g');
            $this->weight = $this->weight * $product['quantity'];
            $price = $_product->get_price() ? $_product->get_price() : 0;
            $price = $price * $product['quantity'];
            $this->orderAmount = +$price;
        }

        return $this;
    }


    /**
     * @return $this
     */
    public function calculate_total_shipping_amount($delivery_options, $signature, $cover)
    {

        foreach ($delivery_options as $option) {
            foreach ($this->categories as $category) {
                if (!($this->weight >= $category['min_weight'] && $this->weight < $category['max_weight'])) {
                   continue;
                }

	            if ($option === 'standard') {
		            $this->shippingAmount[$option . '_parcel'] = $category['standard_price'];
		            $this->shippingAmount[$option . '_collection'] = $category['standard_collection_price'];

		            if ($cover === 'yes') {
			            if ($this->orderAmount > 50 && $signature !== 'yes') {
				            $this->shippingAmount[$option . '_parcel'] += self::SIGNATURE_FEE;
				            $this->shippingAmount[$option . '_collection']  += self::SIGNATURE_FEE;
			            }

			            if ($this->orderAmount <= 999.999 && $this->orderAmount >=  300) {
				            $this->shippingAmount[$option . '_parcel'] += $this->get_cover_fee_for_greater_than_300_and_less_than_1000();
				            $this->shippingAmount[$option . '_collection']  += $this->get_cover_fee_for_greater_than_300_and_less_than_1000();
			            }


			            if ($this->orderAmount <= 299 && $this->orderAmount >=  21) {
				            $this->shippingAmount[$option . '_parcel'] += $this->get_cover_fee();
				            $this->shippingAmount[$option . '_collection']  += $this->get_cover_fee();
			            }
		            }

		            if ($signature == 'yes') {
			            $this->shippingAmount[$option . '_parcel'] += $this->get_signature_fee();
			            $this->shippingAmount[$option . '_collection']  += $this->get_signature_fee();
		            }
	            }

	            if ($option == 'next_day') {
		            $this->shippingAmount[$option] = $category['next_day_price'];

		            if ($cover == 'yes') {
			            if ($this->orderAmount > 50 && $signature !== 'yes') {
				            $this->shippingAmount[$option] += self::SIGNATURE_FEE;
			            }

			            if ($this->orderAmount <= 999.999 && $this->orderAmount >=  300) {
				            $this->shippingAmount[$option]  += $this->get_cover_fee_for_greater_than_300_and_less_than_1000();
			            }

			            if ($this->orderAmount <= 299.99 && $this->orderAmount >=  21) {
				            $this->shippingAmount[$option] += $this->get_cover_fee();
			            }
		            }

		            if ($signature === 'yes') {
			            $this->shippingAmount[$option] += $this->get_signature_fee();
		            }
	            }
            }
        }
        return $this;
    }

    /**
     * @return int
     */
    public function get_shipping_amount()
    {
        return $this->shippingAmount;
    }

    /**
     * @return float
     */
    private function get_signature_fee()
    {
        return self::SIGNATURE_FEE;
    }

    /**
     * @return float
     */
    private function get_cover_fee_for_greater_than_300_and_less_than_1000()
    {

        return $this->cover_rates['greater_than_300_and_less_than_1000'] / 100.0;
    }

    /**
     * @return float
     */
    private function get_cover_fee()
    {
        return ($this->cover_rates[round($this->orderAmount)] / 100.0);
    }
}
