<?php
if ( ! defined( 'ABSPATH' ) ) exit;
if( class_exists('WC_Royal_Mail_Shipping_Method')) return;
/**
 * WC_Royal_Mail_Shipping_Method
 * @author Waseem Senjer
 * @since 1.0.0
 *
 * */
class WC_Royal_Mail_Shipping_Method extends WPRuby_Shipping_Method {

    const PRINTED_PAPER_CLASS = "printed-papers";

    public $supported_services = array(
        'firstclasssmall'					=>	'Standard First Class Small Parcel',
        'firstclassmedium'					=>	'Standard First Class Medium Parcel',
        'secondclasssmallparcel'			=>	'Second Class Small Parcel',
        'secondclassmediumparcel'			=>	'Second Class Medium Parcel',

        'firstclasssignedforsmall'			=>	'Signed For: First Class Small Parcel',
        'firstclasssignedformedium'			=>	'Signed For: First Class Medium Parcel',

        'secondclasssmallparcelsignedfor'	=>	'Signed For: Second Class Small Parcel',
        'secondclassmediumparcelsignedfor'	=>	'Signed For: Second Class Medium Parcel',

        'smallparceltracked24'	=>	'Tracked 24: Small Parcel',
        'mediumparceltracked24'	=>	'Tracked 24: Medium Parcel',

        'smallparceltracked48'	=>	'Tracked 48: Small Parcel',
        'mediumparceltracked48'	=>	'Tracked 48: Medium Parcel',

        'smallparceltracked24withsignature'	=>	'Tracked 24 With Signature: Small Parcel',
        'mediumparceltracked24withsignature'	=>	'Tracked 24 With Signature: Medium Parcel',

        'smallparceltracked48withsignature'	=>	'Tracked 48 With Signature: Small Parcel',
        'mediumparceltracked48withsignature'	=>	'Tracked 48 With Signature: Medium Parcel',
        'specialdelivery9am'				=>	'Special Delivery: Guaranteed by 9am',
        'specialdelivery1pm'				=>	'Special Delivery: Guaranteed by 1pm',
        'specialdelivery1pmsaturday'		=>	'Special Delivery: Guaranteed by 1pm Saturday',
        'specialdelivery9amsaturday'		=>	'Special Delivery: Guaranteed by 9am Saturday',
    );

    public $grouped_services_by_size = [
        'firstclasssmall' => 'first_class',
        'firstclassmedium' => 'first_class',

        'secondclasssmall' => 'second_class',
        'secondclassmedium' => 'second_class',

        'firstclasssigned' => 'first_class_signed',
        'firstclasssignedformedium' => 'first_class_signed',


        'secondclasssmallparcelsignedfor' => 'second_class_signed',
        'secondclassmediumparcelsignedfor' => 'second_class_signed',


        'smallparceltracked24' => 'tracked_24',
        'mediumparceltracked24' => 'tracked_24',

        'smallparceltracked48' => 'tracked_48',
        'mediumparceltracked48' => 'tracked_48',

        'smallparceltracked24withsignature' => 'tracked_24_signed',
        'mediumparceltracked24withsignature' => 'tracked_24_signed',

        'smallparceltracked48withsignature' => 'tracked_48_signed',
        'mediumparceltracked48withsignature' => 'tracked_48_signed',


    ];

    public $supported_international_services = array(
        'internationalstandardsmallparcel'			=>	'International Standard Small Parcel',
        'internationaleconomysmallparcel'			=>	'International Economy Small Parcel',
        'internationalsmallparceltrackedsigned'		=>	'International Tracked and Signed Small Parcel',
        'internationaltrackedsmallparcel'			=>	'International Tracked Small Parcel',
        'internationalsignedsmallparcel'			=>	'International Signed Small Parcel',
        'printedpaperworldwidesigned'              => 'International Signed Printed Papers',
        'printedpaperworldwidestandard'            => 'International Standard Printed Papers',
        'printedpaperworldwidetracked'             => 'International Tracked Printed Papers',
        'printedpaperworldwidetrackedandsigned'    => 'International Tracked & Signed Printed Papers',
    );

    private $intl_printed_papers_options = [
        'printedpaperworldwidesigned',
        'printedpaperworldwidestandard',
        'printedpaperworldwidetracked',
        'printedpaperworldwidetrackedandsigned',
    ];

    private $supported_parcelforce_services = array(
        'parcelforceworldwideexpress9'				=>	'Parcelforce Worldwide: Express 9',
        'parcelforceworldwideexpress10'				=>	'Parcelforce Worldwide: Express 10',
        'parcelforceworldwideexpressAM'				=>	'Parcelforce Worldwide: Express AM',
        'parcelforceworldwideexpress24'				=>	'Parcelforce Worldwide: Express 24',
        'parcelforceworldwideexpress48'				=>	'Parcelforce Worldwide: Express 48',
        'parcelforceworldwideexpress48large'		=>	'Parcelforce Worldwide: Express 48 Large',
        'parcelforceworldwideexpress9saturday'		=>	'Parcelforce Worldwide: Express 9 Saturday',
        'parcelforceworldwideexpress10saturday'		=>	'Parcelforce Worldwide: Express 10 Saturday',
        'parcelforceworldwideexpressAMsaturday'		=>	'Parcelforce Worldwide: Express AM Saturday',
        'parcelforceworldwideexpress24saturday'		=>	'Parcelforce Worldwide: Express 24 Saturday',
        'parcelforceworldwideexpress48saturday'		=>	'Parcelforce Worldwide: Express 48 Saturday',
        'parcelforceworldwideglobaleconomy'			=>	'Parcelforce Worldwide: Global Economy',
        'parcelforceworldwideglobalvalue'			=>	'Parcelforce Worldwide: Global Value',
        'parcelforceworldwideglobalpriority'		=>	'Parcelforce Worldwide: Global Priority',
        'parcelforceworldwideglobalexpress'			=>	'Parcelforce Worldwide: Global Express',
        'parcelforceworldwideeuroprioritybusiness'	=>	'Parcelforce Worldwide: Euro Priority Business',
        'parcelforceworldwideeuropriorityhome'		=>	'Parcelforce Worldwide: Euro Priority Home',
    );
    private $supported_intl_letter_services =
        array(
            'small' =>   array(
                'internationalstandardletter'				=>	'International Standard Letter',
                'internationaleconomyletter'				=>	'International Economy Letter',
                'internationallettertrackedsigned'			=>	'International Tracked and Signed Letter',
                'internationaltrackedletter'				=>	'International Tracked Letter',
                'internationallettersigned'					=>	'International Signed Letter',
            ),
            'large' => array (
                'internationalstandardlargeletter'			=>	'International Standard Large Letter',
                'internationaleconomylargeletter'			=>	'International Economy Large Letter',
                'internationallargelettertrackedandsigned'	=>	'International Tracked and Signed Large Letter',
                'internationaltrackedlargeletter'			=>	'International Tracked Large Letter',
                'internationalsignedlargeletter'			=>	'International Signed Large Letter',
            )
        );


	private $supported_domestic_letter_services = [
	    'small' => [
		    'letter'                            =>	'Standard First Class Letter',
		    'firstclasslettersignedfor'			=>	'Signed For: First Class Letter',
		    'secondclassletter'					=>	'Second Class Letter',
		    'secondclasslettersignedfor'		=>	'Signed For: Second Class Letter',
        ],
	    'large' => [
		    'largeletter'						=>	'Standard First Class Large Letter',
		    'secondclasslargeletter'			=>	'Second Class Large Letter',
		    'firstclasslargelettersignedfor'	=>	'Signed For: First Class Large Letter',
		    'secondclasslargelettersignedfor'	=>	'Signed For: Second Class Large Letter',
		    "largelettertracked24" => "Tracked 24: Large Letter",
            "largelettertracked48" => "Tracked 48: Large Letter",
            "largelettertracked24withsignature" => "Tracked 24 With Signature: Large Letter",
            "largelettertracked48withsignature"  => "Tracked 48 With Signature: Large Letter",
	    ]
	];


    private $handling_fee;
    private $default_weight;
    private $default_size;
    private $parcel_size;

    private $domestic_options;
    private $intl_options;
    private $parcelforce_options;
    private $custom_titles;
    private $custom_rates;
    private $letters_options;
    private $letters_intl_options;
    private $rates_option;
    private $enable_stripping_tax;
    private $enable_letters;
    private $custom_boxes;
    private $with_insurance;

    private $only_letters = true;

    private $package;

    /**
     * WC_Royal_Mail_Shipping_Method constructor.
     *
     * @param int $instance_id
     */
    public function __construct( $instance_id = 0 ) {
        $this->id = 'wpruby_royalmail';
        $this->instance_id = absint( $instance_id );
        $this->supports  = array(
            'shipping-zones',
            'instance-settings',
        );
        $this->method_title = __('Royal Mail Pro', 'wc-royal-mail');
        $this->title = __('Royal Mail', 'wc-royal-mail');

        $this->init_form_fields();
        $this->init_settings();

        $this->tax_status = 'taxable';

        $this->title = $this->get_option('title');
        $this->handling_fee = trim($this->get_option('handling_fee'));

        $this->default_weight = $this->get_option('default_weight');
        $this->default_size = $this->get_option('default_size');

        $this->parcel_size = $this->get_option('parcel_size');

        $this->domestic_options = $this->get_option('domestic_options');
        $this->intl_options = $this->get_option('intl_options');
        $this->parcelforce_options = $this->get_option('parcelforce_options');
        $this->letters_options = $this->get_option('letters_options');
        $this->letters_intl_options = $this->get_option('letters_intl_options');
        $this->rates_option = $this->get_option('rates_option');
        $this->enable_stripping_tax = $this->get_option('enable_stripping_tax');
        $this->custom_titles = $this->get_option('custom_titles');
        $this->custom_rates = $this->get_option('custom_rates');
        $this->custom_boxes = $this->get_option('custom_boxes');

        $this->debug_mode = $this->get_option('debug_mode');
	    $this->with_insurance = $this->get_option('with_insurance');
        $this->enable_letters = $this->get_option('enable_letters');

        add_action('woocommerce_update_options_shipping_' . $this->id, array($this, 'process_admin_options'));
    }

    /**
     *
     */
    public function init_form_fields() {

        $weight_unit = strtolower(get_option('woocommerce_weight_unit'));

        $this->instance_form_fields = array(

            'title' => array(
                'title' => __('Method Title', 'wc-royal-mail'),
                'type' => 'text',
                'description' => __('This controls the title of the method at the checkout.', 'wc-royal-mail'),
                'default' => __('Royal Mail', 'wc-royal-mail'),
                'desc_tip' => true,
            ),
            'debug_mode' => array(
                'title' => __('Enable Debug Mode', 'wc-royal-mail'),
                'type' => 'checkbox',
                'label' => __('Enable ', 'wc-royal-mail'),
                'default' => 'no',
                'description' => __('If debug mode is enabled, the shipping method will be activated just for the administrator. The debug mode will display all the debugging data at the cart and the checkout pages.'),
            ),
            'handling_fee' => array(
                'title' => __('Handling Fees', 'wc-royal-mail'),
                'type' => 'text',
                'css' => 'width:75px',
                'description' => __('(Optional) Enter an amount e.g. 3.5 or a percentage e.g. 3% PS: you can use negative values e.g -3.5', 'wc-royal-mail'),
                'default' => '',
            ),
            'default_weight' => array(
                'title' => __('Default Package Weight', 'wc-royal-mail'),
                'type' => 'text',
                'default' => '0.5',
                'css' => 'width:75px',
                'description' => __("Weight unit: ".$weight_unit."<br> This weight will only be used if the product\s weight are not set in the edit product's page.", 'wc-royal-mail'),
            ),
            'default_size' => array(
                'type' => 'default_size',
            ),
            'parcel_size' => array(
                'title' => __('First / Second Class Parcel Size', 'wc-royal-mail'),
                'type' => 'select',
                'default' => 'small',
                'description' => __('Select the parcel size you\'d like to use for calculating the price of First and Second class parcels. If you select "Small", then the "Small Parcel" price will be used up until 2kg. After 2kg the medium parcel price will always be used.', 'wc-royal-mail'),
                'options' => array(
                    'small' => __('Small Parcel (up to 2kg)', 'wc-royal-mail'),
                    'medium' => __('Medium Parcel', 'wc-royal-mail'),
                ),
            ),
            'domestic_options' => array(
                'title' 	=> __('Domestic Parcel Options', 'wc-royal-mail'),
                'type' 		=> 'multiselect',
                'default' 	=> 'firstclasssmall',
                'class' 	=> 'availability wc-enhanced-select',
                'css' 		=> 'width:80%;',
                'options' 	=> $this->supported_services,
            ),
            'intl_options' => array(
                'title'   => __('International Parcel Options', 'wc-royal-mail'),
                'type' 	  => 'multiselect',
                'css' 	  => 'width:100%;',
                'class'   => 'availability wc-enhanced-select',
                'default' => 'internationalstandardsmallparcel',
                'options' => $this->supported_international_services,
                'description' => __('In order for the Printed Papers options to work, please create a shipping class by the name <span style="padding:1px; background: #c1c1c1; color: #ffffff">'.self::PRINTED_PAPER_CLASS.'</span> and assign it to the paper products.', 'wc-royal-mail'),
            ),
            'parcelforce_options' => array(
                'title' 	=> __('Parcelforce Options', 'wc-royal-mail'),
                'type' 		=> 'multiselect',
                'css' 	  	=> 'width:80%;',
                'class'   	=> 'availability wc-enhanced-select',
                'options' 	=> $this->supported_parcelforce_services,
            ),
            'letters_options' => array(
                'title' 	=> __('Domestic Letters Options', 'wc-royal-mail'),
                'css' 		=> 'width:80%;',
                'type' 		=> 'multiselect',
                'class' 	=> 'availability wc-enhanced-select',
                'options' 	=> $this->supported_domestic_letter_services,
            ),
            'letters_intl_options' => array(
                'title' 	=> __('International Letters Options', 'wc-royal-mail'),
                'css' 		=> 'width:80%;',
                'type' 		=> 'multiselect',
                'class' 	=> 'availability wc-enhanced-select',
                'options' 	=> $this->supported_intl_letter_services,
            ),
            'rates_option' => array(
                'title' 	=> __('Rates ', 'wc-royal-mail'),
                'type' 		=> 'select',
                'default' 	=> 'all',
                'description' => __('Choose whether the plugin shows the cheapest rate or all rates to the customer.', 'wc-royal-mail'),
                'options'  => array(
                    'all' 		=> __('Show All The Options', 'wc-royal-mail'),
                    'cheapest' 	=> __('Show only the cheapest option', 'wc-royal-mail'),
                ),
            ),
            'enable_stripping_tax' => array(
                'title' => __('Remove Tax', 'wc-royal-mail'),
                'type' => 'checkbox',
                'default' => 'no',
                'label' => __('Enable', 'wc-royal-mail'),
                'description' => __('Hint: Enabling this option will strip the tax value (20%) from the shipping prices coming from Royal Mail.', 'wc-royal-mail'),
            ),
            'with_insurance' => array(
	            'title' => __('Consider Insurance', 'wc-royal-mail'),
	            'type' => 'checkbox',
	            'label' => __('Enable ', 'wc-royal-mail'),
	            'default' => 'no',
	            'description' => __('If enabled, the plugin will consider, the amount of the package as a compensation amount, some services might be unavailable if the items in the package has a high price',  'wc-royal-mail'),
            ),
            'custom_titles' => array(
                'type' => 'custom_titles',
                'default' => '',
            ),
            'custom_rates' => array(
                'type' => 'custom_rates',
                'default' => '',
            ),
            'custom_boxes' => array(
                'type' => 'custom_boxes',
                'default' => '',
            ),
        );

    }

    public function calculate_shipping($package = array()) {
        $this->package = $package;
        require_once( ROYALMAILPRO_DIR . 'royalmail/data/Src/CalculateMethod.php' );
        require_once( ROYALMAILPRO_DIR . 'royalmail/data/Src/Data.php' );
        require_once( ROYALMAILPRO_DIR . 'royalmail/data/Src/Method.php' );


        $package_details = $this->get_package_details_by_boxpacker($package);

        if (false === $package_details) {
            $package_details = $this->get_package_details($package);
        }

		$this->debug('Plugin version', ROYALMAILPRO_VERSION);
		$this->debug('Plugin Settings', json_encode($this->instance_settings));
		$this->debug('Packing Details', $package_details);

        $calculateMethodClass = new Meanbee_RoyalmailPHPLibrary_CalculateMethod();
        $dataClass = new Meanbee_RoyalmailPHPLibrary_Data(
            $calculateMethodClass->_csvCountryCode,
            $calculateMethodClass->_csvZoneToDeliverMethod,
            $calculateMethodClass->_csvDeliveryMethodMeta,
            $calculateMethodClass->_csvDeliveryToPrice,
            $calculateMethodClass->_csvCleanNameToMethod,
            $calculateMethodClass->_csvCleanNameMethodGroup
        );

        $rates = [];
        foreach($package_details as $pack){
            $allowedMethods = $this->getAllowedMethods($pack, $package['destination']['country']);
            $this->debug('Allowed Methods', $allowedMethods);

            if (empty($allowedMethods) == true) {
                continue;
            }

            $dataClass->setWeightUnit(strtolower(get_option('woocommerce_weight_unit')));
            $dataClass->_setWeight($pack['weight']);
            $pack_value = (isset($pack['value']))? $pack['value']: 1;
	        $value = ($this->with_insurance === 'yes') ? $pack_value: 1;
            $calculatedMethods = $calculateMethodClass->getMethods(
                $package['destination']['country'],
	            $value,
                (string) round($pack['weight'], 2)
            );

            $this->debug('Calculated Methods',$calculatedMethods);

            if ($this->is_exceeding_limits($pack)) {
                continue;
            }

            if ($this->parcel_size === 'medium' && $this->is_medium_parcel($pack)) {
                $calculatedMethods = $this->filter_methods_by_size($calculatedMethods, ['SMALL']);
            } elseif ($this->is_large_parcel($pack)) {
	            $calculatedMethods = $this->filter_methods_by_size($calculatedMethods, ['SMALL', 'MEDIUM']);
            }elseif($this->is_small_parcel($pack)) {
	            $calculatedMethods = $this->filter_methods_by_size($calculatedMethods, ['MEDIUM']);
            }else{
                $calculatedMethods = $this->filter_methods_by_size($calculatedMethods, ['SMALL']);
            }
	        $allMethods = $this->getAllMethods();

            foreach ($allowedMethods as $allowedMethod) {

                foreach ($calculatedMethods as $methodItem) {
                    if ( ! isset($allMethods[$allowedMethod])) {
                        continue;
                    }

	                if ($allMethods[$allowedMethod] !== $methodItem->shippingMethodNameClean) {
		                continue;
                    }
                    $this->debug('Calculated Method', $methodItem);
                    $price = $methodItem->methodPrice;
                    $price = $this->strip_shipping_tax($price);


	                $rate_id = (isset($this->grouped_services_by_size[$allowedMethod]))? $this->grouped_services_by_size[$allowedMethod]: $methodItem->shippingMethodName;

                    if (!isset($rates[$rate_id])) {
                        $methodTitle = (isset($this->custom_titles[$allowedMethod]) && $this->custom_titles[$allowedMethod] != '') ? $this->custom_titles[$allowedMethod] : $methodItem->shippingMethodNameClean;
                        $rates[$rate_id] = [];
                        $rates[$rate_id]['id'] =  $rate_id;
                        $rates[$rate_id]['label'] = sprintf('%s: %s', $this->title,  $methodTitle) ;
                        $handling_fees = $this->calculate_handling_fee($this->handling_fee, $price);
                        $adjustment = $this->get_adjustment($allowedMethod, $price);
	                    $rates[$rate_id]['cost'] = $price + $handling_fees + $adjustment;

                    }else{
                        $adjustment = $this->get_adjustment($allowedMethod, $price);
                        $rates[$rate_id]['cost'] = $rates[$rate_id]['cost'] + $price + $adjustment;
                    }
                }

            }
        }
	    uasort( $rates, array( $this, 'sort_rates' ) );
	    if($this->rates_option == 'cheapest'){
            // Add the first element which is the cheapest.
            $this->add_rate(reset($rates));
        }else{
	        foreach($rates as $key => $rate){
                $this->add_rate($rate);
            }
        }
    }

    /**
     * get_min_dimension function.
     * get the minimum dimension of the package, so we multiply it with the quantity
     * @access private
     * @param number $width
     * @param number $length
     * @param number $height
     * @return string $result
     */
    public static function get_min_dimension($width, $length, $height) {

        $dimensions = array('width' => $width, 'length' => $length, 'height' => $height);
        $result = array_keys($dimensions, min($dimensions));
        return $result[0];
    }

    /**
     * get_package_details function.
     *
     * @access private
     * @param mixed $package
     * @return array
     */
    private function get_package_details($package) {

        $default_length = (isset($this->default_size['length']))?$this->default_size['length']:0.5;
        $default_width = (isset($this->default_size['width']))?$this->default_size['width']:0.5;
        $default_height = (isset($this->default_size['height']))?$this->default_size['height']:0.5;

        $weight = 0;
        $volume = 0;
        $products = array();
        // Get weight of order
        foreach ($package['contents'] as $item_id => $values) {
            /** @var WC_Product $_product */
            $_product = $values['data'];
            if($_product->is_virtual()){
                continue;
            }
            $final_weight = wc_get_weight((floatval($_product->get_weight()) <= 0) ? $this->default_weight : $_product->get_weight(), 'kg');
            $weight += $final_weight * $values['quantity'];
            $value = $_product->get_price();

            $length = wc_get_dimension((floatval($_product->get_length()) <= 0) ? $default_length : $_product->get_length(), 'cm');
            $height = wc_get_dimension((floatval($_product->get_height()) <= 0) ? $default_height : $_product->get_height(), 'cm');
            $width = wc_get_dimension((floatval($_product->get_width()) <= 0) ? $default_width : $_product->get_width(), 'cm');
            $min_dimension = self::get_min_dimension($width, $length, $height);
            $products[] = array('weight' => wc_get_weight((floatval($_product->get_weight()) <= 0) ? $this->default_weight : $_product->get_weight(), 'kg'),
                'quantity' => $values['quantity'],
                'length' => $length,
                'height' => $height,
                'width' => $width,
                'item_id' => $item_id,
                'value' => $value,
                'min_dimension' => $min_dimension,
            );

			$volume += ($length * $height * $width);
		}
		$max_weights = $this->get_max_weight($products, $package['destination']['country']);
		// @since 1.5 order the products by their postcodes
		usort($products, [$this, 'sort_products_by_length']);

		$pack = array();
		$packs_count = 1;
		$pack[$packs_count]['weight'] = 0;
		$pack[$packs_count]['length'] = 0;
		$pack[$packs_count]['height'] = 0;
		$pack[$packs_count]['width'] = 0;
		$pack[$packs_count]['quantity'] = 0;
		$pack[$packs_count]['value'] = 0;
		$i = 0;
		foreach ($products as $product) {
			$max_weight = $max_weights['own_package'];

			// if only one item is weighing more than the max weight, do not display shipping.
			if ($product['weight'] > $max_weight) {
			    return [];
			}

			if (!$this->is_letter($pack[$packs_count])) {
				$this->only_letters = false;
			}
			while ($product['quantity'] > 0) {
				if (!isset($pack[$packs_count]['weight'])) {
					$pack[$packs_count]['weight'] = 0;
				}
				if (!isset($pack[$packs_count]['quantity'])) {
					$pack[$packs_count]['quantity'] = 0;
				}
				$pack[$packs_count]['weight'] += $product['weight'];
				$pack[$packs_count]['length'] = ('length' == $product['min_dimension']) ? $pack[$packs_count]['length'] + $product['length'] : $product['length'];
				$pack[$packs_count]['height'] = ('height' == $product['min_dimension']) ? $pack[$packs_count]['height'] + $product['height'] : $product['height'];
				$pack[$packs_count]['width'] = ('width' == $product['min_dimension']) ? $pack[$packs_count]['width'] + $product['width'] : $product['width'];
				$pack[$packs_count]['item_id'] = $product['item_id'];
				$pack[$packs_count]['quantity'] += 1;
				$pack[$packs_count]['value'] += round($product['value'], 2);
				if (!$this->is_letter($pack[$packs_count])) {
					$this->only_letters = false;
				}
				if (strval($pack[$packs_count]['weight']) > strval($max_weight)) {

                    $pack[$packs_count]['value'] -= round($product['value'], 2);

                    $pack[$packs_count]['length'] = ('length' == $product['min_dimension']) ? $pack[$packs_count]['length'] - $product['length'] : $product['length'];
                    $pack[$packs_count]['height'] = ('height' == $product['min_dimension']) ? $pack[$packs_count]['height'] - $product['height'] : $product['height'];
                    $pack[$packs_count]['width'] = ('width' == $product['min_dimension']) ? $pack[$packs_count]['width'] - $product['width'] : $product['width'];

                    $pack[$packs_count]['quantity'] -= 1;
                    $pack[$packs_count]['weight'] -= $product['weight'];

                    if(!$this->is_letter($pack[$packs_count])){
                        $this->only_letters = false;
                    }

                    $packs_count++;
                    $pack[$packs_count]['weight'] = $product['weight'];
                    $pack[$packs_count]['length'] = $product['length'];
                    $pack[$packs_count]['height'] = $product['height'];
                    $pack[$packs_count]['width'] = $product['width'];
                    $pack[$packs_count]['item_id'] = $product['item_id'];
                    $pack[$packs_count]['quantity'] = 1;
                    $pack[$packs_count]['value'] = round($product['value'], 2);

                }
                $product['quantity']--;
            }
            $i++;
        }
        return $pack;
    }

    private function is_letter($pack) {

        $slug = "royalmail_not_letter";
        if(isset($this->package['contents'])){
            foreach ($this->package['contents'] as $item_id => $values) {
                /** @var WC_Product $_product */
                $_product = $values['data'];
                $terms = get_the_terms($_product->get_id(), 'product_shipping_class');
                if(!$terms) continue;
                foreach ($terms as $term) {
                    $_shippingclass = $term->slug;
                    if ($slug === $_shippingclass) {
                        return false;
                    }
                }
            }
        }

        if( $pack['weight'] > 0.75 ) {
            return false;
        }

        if (wc_get_dimension($pack['length'], 'cm') > 35.3) {
            return false;
        }

        if (wc_get_dimension($pack['height'], 'cm') > 2.5) {
            return false;
        }

        return true;
    }

    private function is_small_letter($pack, $country = 'GB'){

        if ($this->is_cart_has_no_letter_shipping_class()) {
            return false;
        }
	    $dimensions = $this->sorted_dimensions($pack);

	    $is_small_letter = ($pack['weight'] <= 0.1  && $dimensions['length'] <= 24 && $dimensions['height'] <= 0.5);

        if ('GB' != $country) {
            return $is_small_letter;
        }
        $is_enabled_small_letters = !empty(array_intersect(array_keys($this->supported_domestic_letter_services['small']), $this->letters_options));

        return $is_small_letter && $is_enabled_small_letters;
    }


    private function is_large_letter($pack, $country = 'GB'){
        if ($this->is_cart_has_no_letter_shipping_class()) {
            return false;
        }

        $dimensions = $this->sorted_dimensions($pack);
		$is_large_letter = ($pack['weight'] <= 0.75  && intval($dimensions['length'] *10)<= 353 && $dimensions['width'] <= 25 && $dimensions['height'] <= 2.5);


        if ('GB' != $country) {
            return $is_large_letter;
        }

        $is_enabled_large_letters = !empty(array_intersect(array_keys($this->supported_domestic_letter_services['large']), $this->letters_options));

        return $is_large_letter && $is_enabled_large_letters;
    }

    /**
     * @return bool
     */
    private function is_cart_has_no_letter_shipping_class(){
        $slug = "royalmail_not_letter";
        if(isset($this->package['contents'])){
            foreach ($this->package['contents'] as $item_id => $values) {
                /** @var WC_Product $_product */
                $_product = $values['data'];
                $terms = get_the_terms($_product->get_id(), 'product_shipping_class');

                if ($terms) {
                    foreach ($terms as $term) {
                        $_shipping_class = $term->slug;
                        if ($slug === $_shipping_class) {
                            return true;
                        }
                    }
                }
            }
        }

        return false;
    }

	/**
	 * @param $slug
	 *
	 * @return bool
	 */
    private function is_every_item_in_cart_cart_has_specified_shipping_class($slug)
    {
        $found = false;

        if (!isset($this->package['contents'])) {
            return $found;
        }


        foreach ($this->package['contents'] as $item_id => $values) {
            $found = true;
            /** @var WC_Product $_product */
            $_product = $values['data'];
            $terms = get_the_terms($_product->get_id(), 'product_shipping_class');

            if (!is_wp_error($terms) && $terms) {
                foreach ($terms as $term) {
                    if ($slug !== $term->slug) {
                        return false;
                    }
                }
            } else {

                return false;
            }
        }

            return $found;


    }

    /**
     * @param $products
     * @param $country
     *
     * @return array
     */
    private function get_max_weight($products, $country) {

        $pack = array();
        $pack['length'] = 0;
        $pack['width'] = 0;
        $pack['height'] = 0;
        $pack['weight'] = 0;

        foreach ($products as $product) {
            while ( $product['quantity'] > 0 ) {
                $pack['length'] += $product['length'];
                $pack['width'] +=  $product['width'];
                $pack['height'] += $product['height'];
                $pack['weight'] += $product['weight'];
                $product['quantity']--;
            }
        }

        if ($country != 'GB') {
	        $all_printed = $this->is_every_item_in_cart_cart_has_specified_shipping_class(self::PRINTED_PAPER_CLASS);
	        if ($this->is_printed_paper_enabled() && $this->parcel_size == 'small' && $all_printed === true) {
		        return array('own_package' => 5);
	        }

	        return array('own_package' => 30);
        }

        if ($this->is_small_parcel($pack) && $this->parcel_size == 'small') {
            return array('own_package' => 2);
        }

        return array('own_package' => 20);
    }


    /**
     * @return string
     */
    public function generate_default_size_html()
    {
        $dimensions_unit = strtolower(get_option('woocommerce_dimension_unit'));
        if(version_compare(WC()->version, '2.6.0', 'lt')){
            $length = (isset($this->default_size['length']))?$this->default_size['length']:'';
            $width  = (isset($this->default_size['width']))?$this->default_size['width']:'';
            $height  = (isset($this->default_size['height']))?$this->default_size['height']:'';
        }else{
            $length = (isset($this->instance_settings['default_size']['length']))?$this->instance_settings['default_size']['length']:'';
            $width = (isset($this->instance_settings['default_size']['width']))?$this->instance_settings['default_size']['width']:'';
            $height = (isset($this->instance_settings['default_size']['height']))?$this->instance_settings['default_size']['height']:'';
        }
        ob_start();
        ?>
        <tr valign="top">
            <th scope="row" class="titledesc">
                <label><?php _e('Default Package Size', 'wc-royal-mail') ?></label>
            </th>
            <td class="forminp">
                <fieldset>
                    <label for="woocommerce_wpruby_royalmail_default_length"><?php _e('Length', 'wc-royal-mail'); ?></label> <input id="woocommerce_wpruby_royalmail_default_length" name="woocommerce_wpruby_royalmail_default_length" type="text" value="<?php echo esc_attr($length); ?>" style="width:70px" />
                    <label for="woocommerce_wpruby_royalmail_default_width"><?php _e('Width', 'wc-royal-mail'); ?></label>  <input id="woocommerce_wpruby_royalmail_default_width" name="woocommerce_wpruby_royalmail_default_width" type="text" value="<?php echo esc_attr($width); ?>" style="width:70px" />
                    <label for="woocommerce_wpruby_royalmail_default_height"><?php _e('Height', 'wc-royal-mail'); ?></label> <input id="woocommerce_wpruby_royalmail_default_height" name="woocommerce_wpruby_royalmail_default_height" type="text" value="<?php echo esc_attr($height); ?>" style="width:70px" />
                    <p class="description">Size unit: <?php echo $dimensions_unit; ?><br> This dimension will only be used if the product\s dimensions are not set in the edit product's page.</p>
                </fieldset>
            </td>
        </tr>

        <?php
        return ob_get_clean();

    }

    /**
     * validate_default_size_field function.
     *
     * @access public
     *
     * @param mixed $key
     *
     * @return array
     */
    public function validate_default_size_field($key) {
        $dimensions = array();
        if(is_numeric($_POST['woocommerce_wpruby_royalmail_default_length']) && $_POST['woocommerce_wpruby_royalmail_default_length'] > 0)
            $dimensions['length'] = $_POST['woocommerce_wpruby_royalmail_default_length'];
        if(is_numeric($_POST['woocommerce_wpruby_royalmail_default_width']) && $_POST['woocommerce_wpruby_royalmail_default_width'] > 0)
            $dimensions['width']  = $_POST['woocommerce_wpruby_royalmail_default_width'];
        if(is_numeric($_POST['woocommerce_wpruby_royalmail_default_height']) && $_POST['woocommerce_wpruby_royalmail_default_height'] > 0)
            $dimensions['height'] = $_POST['woocommerce_wpruby_royalmail_default_height'];

        return $dimensions;
    }

    private function getAllowedMethods($pack, $country){

        $letters_options = (($this->letters_options == ''))?array():(((is_array($this->letters_options)))?$this->letters_options:array($this->letters_options));
        $letters_intl_options = (($this->letters_intl_options == ''))?array():(((is_array($this->letters_intl_options)))?$this->letters_intl_options:array($this->letters_intl_options));
        $domestic_options = (($this->domestic_options == ''))?array():(((is_array($this->domestic_options)))?$this->domestic_options:array($this->domestic_options));
        $intl_options = (($this->intl_options == ''))?array():(((is_array($this->intl_options)))?$this->intl_options:array($this->intl_options));
        $parcelforce_options = (($this->parcelforce_options == ''))?array():(((is_array($this->parcelforce_options)))?$this->parcelforce_options:array($this->parcelforce_options));


        $remove_all_other_options = false;

        if ($intl_options) {
            $all_printed = $this->is_every_item_in_cart_cart_has_specified_shipping_class(self::PRINTED_PAPER_CLASS);
            $options = $intl_options;
            if (!$all_printed) {
                foreach ($options as $key => $option) {
	                if (in_array($option, $this->intl_printed_papers_options)){
                        unset($intl_options[$key]);
                    }
                }
            } else {
                foreach ($this->intl_printed_papers_options as $value) {
                        if (in_array($value, $intl_options)) {
                            $remove_all_other_options = true;
                            foreach ($options as $key => $option) {
                                if(!in_array($option, $this->intl_printed_papers_options)){
                                    unset($intl_options[$key]);
                                }
                            }
                            break;
                        }
                }
            }
        }

        if ($this->letters_options && 'GB' === $country ){
            // return the letters option only if the package weigh less than 750g.
            if ($this->only_letters === true && ($this->is_large_letter($pack, $country) || $this->is_small_letter($pack, $country))) {
                if ($this->is_small_letter($pack, $country)){
                    $small_letters_options = array_intersect($letters_options, array_keys($this->supported_domestic_letter_services['small']));
                    if (!empty($small_letters_options)) {
                        return $small_letters_options;
                    }
                }

                if ($this->is_large_letter($pack, $country)) {
                    $large_letters_options = array_intersect($letters_options, array_keys($this->supported_domestic_letter_services['large']));
                    if (!empty($large_letters_options)) {
                        return $large_letters_options;
                    }
                }

                return $letters_options;
            }else{
                if (!$this->is_large_parcel($pack) && !$this->is_medium_parcel($pack) && !$this->is_small_parcel($pack)) {
                    return $parcelforce_options;
                }

                return  array_merge($domestic_options, $parcelforce_options);
            }
        }elseif($this->letters_intl_options && 'GB' !== $country){
            // return the letters option only if the package weigh less than 750g.
            if($this->is_letter($pack) && $this->only_letters === true){
                if ($this->is_small_letter($pack, $country)){
                    $small_letters_options = array_intersect($letters_intl_options, array_keys($this->supported_intl_letter_services['small']));
                    if (!empty($small_letters_options)) {
                        return $small_letters_options;
                    }
                }

                if ($this->is_large_letter($pack, $country)) {
                    $large_letters_options = array_intersect($letters_intl_options, array_keys($this->supported_intl_letter_services['large']));
                    if (!empty($large_letters_options)) {
                        return $large_letters_options;
                    }
                }
            }else{
                return array_merge($intl_options, $parcelforce_options);
            }
        }else{
            return $remove_all_other_options ? array_merge($intl_options, $parcelforce_options) : array_merge($domestic_options, $intl_options, $parcelforce_options);
        }

	    return array_merge($intl_options, $parcelforce_options);
    }

    private function getAllMethods(){
        return array_merge(
            $this->supported_services,
            $this->supported_international_services,
            $this->supported_parcelforce_services,
            $this->supported_intl_letter_services['small'],
            $this->supported_intl_letter_services['large'],
            $this->supported_domestic_letter_services['small'],
            $this->supported_domestic_letter_services['large']
        );
    }

    /** sort rates based on cost **/
    public function sort_rates( $a, $b ) {
        if ( $a['cost'] == $b['cost'] ) return 0;
        return ( $a['cost'] < $b['cost'] ) ? -1 : 1;
    }

    /**
     * @since 1.8.6
     * Generates the HTML for custom shipping methods label
     **/
    public function generate_custom_titles_html()
    {
        if (version_compare(WC()->version, '2.6.0', 'lt')) {
            $custom_titles = (isset($this->custom_titles))?$this->custom_titles:'';
            $custom_rates = (isset($this->custom_rates))?$this->custom_rates:'';
        } else {
            $custom_titles = (isset($this->instance_settings['custom_titles']))?$this->instance_settings['custom_titles']:'';
            $custom_rates = (isset($this->instance_settings['custom_rates']))?$this->instance_settings['custom_rates']:'';
        }
        ob_start();
        require_once( ROYALMAILPRO_DIR . 'shared/views/custom_titles_html.php' );
        return ob_get_clean();
    }

    /**
     * @since 1.8.6
     * Generates the HTML for custom shipping methods label
     **/
    public function generate_custom_rates_html()
    {

    }

    /**
     * validate_custom_titles_field function.
     *
     * @access public
     * @return array
     * @internal param mixed $key
     */
    public function validate_custom_titles_field()
    {
        $custom_titles = array();
        $posted_custom_titles = array();
        if (isset($_POST['woocommerce_custom_titles'])) {
            $posted_custom_titles = $_POST['woocommerce_custom_titles'];
        }
        if (!empty($posted_custom_titles)) {
            foreach ($posted_custom_titles as $type => $value) {
                $custom_titles[$type] = $value;
            }
        }
        return $custom_titles;
    }

    /**
     * validate_custom_rates_field function.
     *
     * @access public
     * @return array
     * @internal param mixed $key
     */
    public function validate_custom_rates_field()
    {
        $custom_rates = array();
        $posted_custom_rates = array();
        if (isset($_POST['woocommerce_custom_rates'])) {
            $posted_custom_rates = $_POST['woocommerce_custom_rates'];
        }
        if (!empty($posted_custom_rates)) {
            foreach ($posted_custom_rates as $type => $value) {
                $custom_rates[$type] = $value;
            }
        }
        return $custom_rates;
    }

    private function is_small_parcel($params)
    {
        $dimensions = $this->sorted_dimensions($params);
        if ($params['weight'] > 2){
            return false;
        }

        if ($dimensions['length'] > 45){
            return false;
        }

        if ($dimensions['width'] > 35){
            return false;
        }

        if ($dimensions['height'] > 16) {
            return false;
        }

        return true;
    }

    private function is_medium_parcel($dimensions){
        if ($dimensions['weight'] > 20) {
            return false;
        }

        if ($dimensions['length'] > 61){
            return false;
        }

        if ($dimensions['width'] > 46) {
            return false;
        }

        if ($dimensions['height'] > 46) {
            return false;
        }

        return !$this->is_small_parcel($dimensions);
    }

    /**
     * @param string $rate_key
     *
     * @param $total
     *
     * @return int
     */
    private function get_adjustment($rate_key, $total){
        if(!isset($this->custom_rates[$rate_key]) || $this->custom_rates[$rate_key] == ''){
            return 0;
        }

        if (substr($this->custom_rates[$rate_key], -1) == '%') {
            $adjustment = trim(str_replace('%', '', $this->custom_rates[$rate_key]));
            return ($total * ($adjustment / 100));
        }

        if (is_numeric($this->custom_rates[$rate_key])) {
            return ($this->custom_rates[$rate_key]);
        }
        return 0;

    }

    /**
     * @param $methods
     * @param array $sizes
     *
     * @return mixed
     */
    public function filter_methods_by_size($methods, $sizes){
        foreach ($methods as $key => $value) {
            if ( in_array($value->size, $sizes) ) {
                unset($methods[$key]);
            }
        }
        return $methods;
    }

    /**
     * @param $rate_cost
     *
     * @return float
     */
    private function strip_shipping_tax($rate_cost){
        if('yes' !== $this->enable_stripping_tax){
            return $rate_cost;
        }

        return $rate_cost / 1.2;
    }

    /**
     * @since 2.0.0
     * Generates the HTML for custom shipping methods label
     **/
    public function generate_custom_boxes_html()
    {
        ob_start();
        require_once( ROYALMAILPRO_DIR . 'shared/views/custom_boxes_html.php' );
        return ob_get_clean();
    }


    /**
     * validate_custom_boxes_field function.
     *
     * @access public
     * @return array
     * @internal param mixed $key
     */
    public function validate_custom_boxes_field()
    {
        $custom_boxes = array();
        $posted_custom_boxes = array();
        if (isset($_POST['woocommerce_custom_boxes'])) {
            $posted_custom_boxes = $_POST['woocommerce_custom_boxes'];
        }

        if (!empty($posted_custom_boxes)) {
            foreach ($posted_custom_boxes as $key => $value) {
                foreach ($value as $i => $attr) {
                    $custom_boxes[$i][$key] = $attr;
                }
            }
        }
        return $custom_boxes;
    }


    /**
     * get_package_details_by_boxpacker function.
     *
     * @access private
     * @param mixed $package
     * @return mixed
     */
    private function get_package_details_by_boxpacker($package)
    {
        require_once ROYALMAILPRO_DIR . 'shared/boxpacker/class-packer.php';

        $default_length = isset($this->default_size['length'])?$this->default_size['length']:1;
        $default_width 	=  isset($this->default_size['width'])?$this->default_size['width']:1;
        $default_height = isset($this->default_size['height'])?$this->default_size['height']:1;
        //1. adding boxes
        $boxes = array();

        if (!empty($this->custom_boxes)) {
            foreach ($this->custom_boxes as $key => $box) {
                $boxes[] = new WPRuby_RoyalMailBox($box['box_name'], $box['box_outer_length'], $box['box_outer_width'], $box['box_outer_height'], $box['box_empty_weight'], $box['box_inner_length'], $box['box_inner_width'], $box['box_inner_height'], $box['box_maximum_weight'] +  $box['box_empty_weight']);
            }
        } else {
            $boxes = $this->add_parcels_as_boxes();
        }

        $packer = new WPRuby_Packer();
        foreach ($boxes as $box) {
            $packer->addBox($box);
        }

        // Get weight of order
        foreach ($package['contents'] as $item_id => $values) {
            /** @var WC_Product $_product */
            $_product = $values['data'];
            //info: since 1.9.2 skipp virtual products
            if ($_product->is_virtual()) {
                continue;
            }

            $weight = wc_get_weight((floatval($_product->get_weight()) <= 0) ? $this->default_weight : $_product->get_weight(), 'g');
            $value = $_product->get_price();

            $length = wc_get_dimension((floatval($_product->get_length()) <= 0) ? $default_length : $_product->get_length(), 'mm');
            $height = wc_get_dimension((floatval($_product->get_height()) <= 0) ? $default_height : $_product->get_height(), 'mm');
            $width = wc_get_dimension((floatval($_product->get_width()) <= 0) ? $default_width : $_product->get_width(), 'mm');
            //adding the packer code
            //2. adding items
            for ($i = 0 ; $i < $values['quantity']; $i++) {
                $item = new WPRuby_RoyalMailItem("Product", $width, $length, $height, $weight,  false, $value);
                $packer->addItem($item);
            }
            //end of the packer code
        }

        //adding the packer code
        //3. packing
        try {
            $packedBoxes = $packer->pack();
        } catch (Exception $e) {
            return false;
        }

        $pack = array();
        $packs_count = 1;
        $pack[$packs_count]['weight'] = 0;
        $pack[$packs_count]['length'] = 0;
        $pack[$packs_count]['height'] = 0;
        $pack[$packs_count]['width'] = 0;
        $pack[$packs_count]['quantity'] = 0;
        $pack[$packs_count]['value'] = 0;
        /** @var WPRuby_PackedBox $packedBox */
        foreach ($packedBoxes as $packedBox) {
            /** @var WPRuby_RoyalMailBox $boxType */
            $pack[$packs_count]['weight'] = $packedBox->getWeight() / 1000;
            $pack[$packs_count]['length'] = $packedBox->getUsedLength() / 10;
            $pack[$packs_count]['width'] =  $packedBox->getUsedWidth() / 10;
	        // TODO: check why in some cases the usedDepth is zero.
	        $pack[$packs_count]['height'] = ($packedBox->getUsedDepth() === 0)? $height / 10: $packedBox->getUsedDepth() / 10;
            $pack[$packs_count]['quantity'] = count($packedBox->getItems()->asArray());
            $pack[$packs_count]['postcode'] = $package['destination']['postcode'];
            $pack[$packs_count]['value'] = array_reduce($packedBox->getItems()->asArray(), function ($carry, $item){
                $carry += $item->getPrice();
                return $carry;
            });
            $packs_count++;
        }
        return $pack;
    }

    private function add_parcels_as_boxes()
    {

        $boxes = [];

	    if ($this->is_small_letters_enabled()) {
		    $boxes[] = new WPRuby_RoyalMailBox(
			    'Small Letter',
			    240,
			    165,
			    5,
			    0,
			    240,
			    165,
			    5,
			    100
		    );
	    }

        if ($this->is_large_letters_enabled()) {
	        $boxes[] = new WPRuby_RoyalMailBox(
		        'Large Letter',
		        353,
		        250,
		        25,
		        0,
		        350,
		        255,
		        24,
		        750
	        );
        }

        if ($this->parcel_size === 'small'){
            $boxes[] = new WPRuby_RoyalMailBox(
                'Small Parcel',
                450,
                350,
                160,
                0,
	            450,
	            350,
	            160,
                2000
            );
        }

        if ($this->parcel_size === 'medium') {
	        $boxes[] = new WPRuby_RoyalMailBox(
		        'Small Parcel',
		        450,
		        350,
		        160,
		        0,
		        450,
		        350,
		        160,
		        2000
	        );
            $boxes[] = new WPRuby_RoyalMailBox(
                'Medium Parcel',
                610,
                460,
                460,
                0,
	            610,
	            460,
	            460,
                20 * 1000
            );
            return $boxes;
        }

        return [];
    }

    /**
     * @param $pack
     *
     * @return bool
     */
    private function is_large_parcel($pack) {
        $not_small_or_medium = !$this->is_medium_parcel($pack) && !$this->is_small_parcel($pack);
        $girth = $pack['length'] + (2 * $pack['height'] + 2 * $pack['width']);
        return $not_small_or_medium && ($girth < 300);
    }

	private function is_exceeding_limits($pack)
    {
        // 4m is the max length of ParcelForce Large
        if ($pack['length'] > 400) {
            return true;
        }

	    if ($pack['width'] > 400) {
		    return true;
	    }

	    if ($pack['height'] > 400) {
		    return true;
	    }

		return false;
	}

	/**
	 * @return bool
	 */
	private function is_printed_paper_enabled()
    {
        if (!is_array($this->intl_options)) {
            return false;
        }

        if (count($this->intl_options) === 0) {
            return false;
        }

        return !empty(array_intersect($this->intl_options, $this->intl_printed_papers_options));
    }

	private function sort_products_by_length($productA, $productB)
	{
		return $productA['length'] - $productB['length'];
	}

	private function sorted_dimensions($pack)
    {
        $dimensions = array(
          $pack['length'],
          $pack['width'],
          $pack['height'],
        );

        rsort($dimensions);

        return array(
          'length' => $dimensions[0],
          'width' => $dimensions[1],
          'height' => $dimensions[2],
        );
    }

	/**
	 * @return bool
	 */
	private function is_large_letters_enabled()
    {
        $letters_options = (is_array($this->letters_options))? $this->letters_options: [$this->letters_options];
        return count(array_intersect($letters_options, array_keys($this->supported_domestic_letter_services['large']))) > 0;
    }

	private function is_small_letters_enabled()
	{
		$letters_options = (is_array($this->letters_options))? $this->letters_options: [$this->letters_options];
		return count(array_intersect($letters_options, array_keys($this->supported_domestic_letter_services['small']))) > 0;
	}
}
