<?php

abstract class WPRuby_Abstract_Calculator {

	protected $settings = [];

	/**
	 * get_package_details_by_boxpacker function.
	 *
	 * @access private
	 * @param mixed $package
	 * @return mixed
	 */
	protected function get_package_details_by_boxpacker($package)
	{
		require_once ROYALMAILPRO_DIR. 'shared/boxpacker/class-packer.php';

		$default_length = isset($this->settings['default_size']['length'])?$this->settings['default_size']['length']:10;
		$default_width 	=  isset($this->settings['default_size']['width'])?$this->settings['default_size']['width']:10;
		$default_height = isset($this->settings['default_size']['height'])?$this->settings['default_size']['height']:10;

		$packer = new WPRuby_Packer();
		//1. adding boxes
		foreach ($this->add_boxes_iterations() as $box) {
			$packer->addBox($box);
		}

		// Get weight of order
		foreach ($package['contents'] as $item_id => $values) {
			/** @var WC_Product $_product */
			$_product = $values['data'];
			//info: since 1.9.2 skipp virtual products
			if ($_product->is_virtual()) {
				continue;
			}

			$weight = (floatval($_product->get_weight()) <= 0) ? $this->settings['weight'] : wc_get_weight($_product->get_weight(), 'g');
			$value = $_product->get_price();

			$length = (floatval($_product->get_length()) <= 0) ? $default_length : wc_get_dimension($_product->get_length(), 'mm');
			$height = (floatval($_product->get_height()) <= 0) ? $default_height : wc_get_dimension($_product->get_height(), 'mm');
			$width = (floatval($_product->get_width()) <= 0) ? $default_width : wc_get_dimension($_product->get_width(), 'mm');
			//adding the packer code
			//2. adding items
			for ($i = 0 ; $i < $values['quantity']; $i++) {
				$item = new WPRuby_RoyalMailItem("Product", $width, $length, $height, $weight,  false, $value);
				$packer->addItem($item);
			}
			//end of the packer code
		}

		//adding the packer code
		//3. packing
		try {
			$packedBoxes = $packer->pack();
		} catch (Exception $e) {
			return false;
		}

		$pack = array();
		$packs_count = 1;
		$pack[$packs_count]['weight'] = 0;
		$pack[$packs_count]['length'] = 0;
		$pack[$packs_count]['height'] = 0;
		$pack[$packs_count]['width'] = 0;
		$pack[$packs_count]['quantity'] = 0;
		$pack[$packs_count]['value'] = 0;
		/** @var WPRuby_PackedBox $packedBox */
		foreach ($packedBoxes as $packedBox) {
			/** @var WPRuby_RoyalMailBox $boxType */
			$pack[$packs_count]['weight'] = $packedBox->getWeight() / 1000;
			$pack[$packs_count]['length'] = $packedBox->getUsedLength() / 10;
			$pack[$packs_count]['width'] =  $packedBox->getUsedWidth() / 10;
			$pack[$packs_count]['height'] = ($packedBox->getUsedDepth() === 0)? $height / 10: $packedBox->getUsedDepth() / 10;
			$pack[$packs_count]['quantity'] = count($packedBox->getItems()->asArray());
			$pack[$packs_count]['postcode'] = $package['destination']['postcode'];
			$pack[$packs_count]['value'] = array_reduce($packedBox->getItems()->asArray(), function ($carry, $item){
				$carry += $item->getPrice();
				return $carry;
			});
			$packs_count++;
		}
		return $pack;
	}

	 /**
	 * @return array
	 */
	protected function add_boxes_iterations()
	{
		$boxes = [];
		$length = 100;
		$width = 50;
		$height = 20;
		$emptyWeight = 50;
		$maxWeight = 1000;

		foreach (range(1, 15) as $box) {
			$boxes[] = new WPRuby_RoyalMailBox( 'Box #' . $box, $length, $width, $height, $emptyWeight, $length - 10, $width - 10, $height - 10, $maxWeight );
			$length += 200;
			$width += 100;
			$height += 30;
			$emptyWeight += 50;
			$maxWeight += 1000;
		}

		return $boxes;
	}

	protected function normalizeCountryName($country)
	{
		$countryParts = explode('(', $country);
		return trim($countryParts[0]);
	}

	/**
	 * Output a message
	 *
	 * @param $message
	 * @param $array
	 * @param string $type
	 */
	protected function debug($message, $array, $type = 'notice') {
		if ($this->settings['debug_mode'] == 'yes' && current_user_can('manage_options')) {
			$message = $message . ': <pre>' . json_encode($array, JSON_PRETTY_PRINT). '</pre>';
			wc_add_notice($message, $type);
		}
	}
}
